@@ -1,19 +1,19 @@
 package PDF::FromHTML;
-$PDF::FromHTML::VERSION = '0.04';
+$PDF::FromHTML::VERSION = '0.05';
 
 use strict;
 use warnings;
-use Spiffy '-base';
 
-field 'pdf';
-field 'twig';
-field 'args';
+BEGIN {
+    foreach my $method ( qw( pdf twig tidy args ) ) {
+        no strict 'refs';
+        *$method = sub { $#_ ? ($_[0]{$method} = $_[1]) : $_[0]{$method} };
+    }
+}
 
 use Cwd;
-use XML::Clean;
 use File::Temp;
 use File::Basename;
-use Hook::LexWrap;
 
 use PDF::Writer 'pdfapi2';
 use PDF::Template;
@@ -25,8 +25,8 @@ PDF::FromHTML - Convert HTML documents to PDF
 
 =head1 VERSION
 
-This document describes version 0.04 of PDF::FromHTML, released 
-September 23, 2004.
+This document describes version 0.05 of PDF::FromHTML, released 
+November 18, 2004.
 
 =head1 SYNOPSIS
 
@@ -55,6 +55,7 @@ sub new {
     my $class = shift;
     bless({
         twig => PDF::FromHTML::Twig->new,
+        tidy => eval { require HTML::Tidy; HTML::Tidy->new },
         args => { @_ },
     }, $class);
 }
@@ -67,30 +68,47 @@ sub load_file {
 sub parse_file {
     my $self = shift;
     my $file = $self->{file};
+    my $content = '';
 
     my $dir = Cwd::getcwd();
+
     if (!ref $file) {
         open my $fh, $file or die $!;
         chdir File::Basename::dirname($file);
-        $file = \do { local $/; <$fh> };
+        $content = do { local $/; <$fh> };
     }
     else {
-        $file = \"$$file";
+        $content = $$file;
     }
-    $$file =~ s{<!--\s.*?\s-->}{}gs;
 
-    # lower-case all tags
-    my $lc_tags = Hook::LexWrap::wrap(
-        *XML::Clean::handle_start,
-        pre => sub {
-            $_[0] = lc($_[0]);
-            $_[0] = 'i' if $_[0] eq 'em';
-            $_[0] = 'b' if $_[0] eq 'strong';
-        },
-    );
+    if (my $encoding = ($self->args->{encoding} || 'utf8') and $] >= 5.007003) {
+        require Encode;
+        $content = Encode::decode($encoding, $content, Encode::FB_XMLCREF());
+    }
+
+    $content =~ s{&nbsp;}{}g;
+    $content =~ s{<!--.*?-->}{}gs;
+
+    if (my $tidy = $self->tidy) {
+        if ($] >= 5.007003) {
+            $content = Encode::encode( ascii => $content, Encode::FB_XMLCREF());
+        }
+        $content = $self->tidy->clean(
+            '<?xml version="1.0" encoding="UTF-8"?>',
+            '<html xmlns="http://www.w3.org/1999/xhtml">',
+            $content,
+        );
+    }
+    else {
+        require XML::Clean;
+        $content =~ s{&#(\d+);}{chr $1}eg;
+        $content =~ s{&#x([\da-fA-F]+);}{chr hex $1}eg;
+        $content = XML::Clean::clean($content, '1.0', { encoding => 'UTF-8' });
+        $content =~ s{<(/?\w+)}{<\L$1}g;
+    }
+
+    $self->twig->parse( $content );
 
-    my $cleaned = XML::Clean::clean($$file, '1.0', $self->args);
-    $self->twig->parse($cleaned);
     chdir $dir;
 }
 
@@ -109,10 +127,14 @@ sub convert {
         SUFFIX => '.xml',
         UNLINK => 1,
     );
-    binmode($fh);
+
+    binmode($fh, ($] >= 5.007003) ? (':utf8') : ());
+
+    # print STDERR "==> Temp file written to $filename\n";
     print $fh $self->twig->sprint;
     close $fh;
 
+    local $^W;
     $self->pdf(eval { PDF::Template->new( filename => $filename ) })
       or die "$filename: $@";
     $self->pdf->param(@_);
@@ -120,6 +142,7 @@ sub convert {
 
 sub write_file {
     my $self = shift;
+    local $^W;
     $self->pdf->write_file(@_);
 }
 
